@@ -1,8 +1,8 @@
 /**************************************************************************
  *   nano.c  --  This file is part of GNU nano.                           *
  *                                                                        *
- *   Copyright (C) 1999-2011, 2013-2021 Free Software Foundation, Inc.    *
- *   Copyright (C) 2014-2020 Benno Schulenberg                            *
+ *   Copyright (C) 1999-2011, 2013-2022 Free Software Foundation, Inc.    *
+ *   Copyright (C) 2014-2021 Benno Schulenberg                            *
  *                                                                        *
  *   GNU nano is free software: you can redistribute it and/or modify     *
  *   it under the terms of the GNU General Public License as published    *
@@ -26,7 +26,7 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <getopt.h>
-#if defined(__linux__) || !defined(NANO_TINY)
+#ifdef __linux__
 #include <sys/ioctl.h>
 #endif
 #ifdef ENABLE_UTF8
@@ -76,7 +76,6 @@ linestruct *make_new_node(linestruct *prevnode)
 #endif
 	newnode->lineno = (prevnode) ? prevnode->lineno + 1 : 1;
 #ifndef NANO_TINY
-	newnode->extrarows = -2;  /* Bad value, to make it easier to find bugs. */
 	newnode->has_anchor = FALSE;
 #endif
 
@@ -100,6 +99,9 @@ void splice_node(linestruct *afterthis, linestruct *newnode)
 /* Free the data structures in the given node. */
 void delete_node(linestruct *line)
 {
+	/* If the first line on the screen gets deleted, step one back. */
+	if (line == openfile->edittop)
+		openfile->edittop = line->prev;
 #ifdef ENABLE_WRAPPING
 	/* If the spill-over line for hard-wrapping is deleted... */
 	if (line == openfile->spillage_line)
@@ -152,8 +154,7 @@ linestruct *copy_node(const linestruct *src)
 #endif
 	dst->lineno = src->lineno;
 #ifndef NANO_TINY
-	dst->extrarows = src->extrarows;
-	dst->has_anchor = FALSE;
+	dst->has_anchor = src->has_anchor;
 #endif
 
 	return dst;
@@ -211,6 +212,18 @@ bool in_restricted_mode(void)
 		return FALSE;
 }
 
+#ifndef NANO_TINY
+/* Say how the user can achieve suspension (when they typed ^Z). */
+void suggest_ctrlT_ctrlZ(void)
+{
+#ifdef ENABLE_NANORC
+	if (first_sc_for(MMAIN, do_execute) && first_sc_for(MMAIN, do_execute)->keycode == 0x14 &&
+			first_sc_for(MEXECUTE, do_suspend) && first_sc_for(MEXECUTE, do_suspend)->keycode == 0x1A)
+#endif
+		statusline(AHEM, _("To suspend, type ^T^Z"));
+}
+#endif
+
 /* Make sure the cursor is visible, then exit from curses mode, disable
  * bracketed-paste mode, and restore the original terminal settings. */
 void restore_terminal(void)
@@ -230,14 +243,14 @@ void finish(void)
 	/* Blank the status bar and (if applicable) the shortcut list. */
 	blank_statusbar();
 	blank_bottombars();
-	wrefresh(bottomwin);
+	wrefresh(footwin);
 
 #ifndef NANO_TINY
 	/* Deallocate the two or three subwindows. */
 	if (topwin != NULL)
 		delwin(topwin);
-	delwin(edit);
-	delwin(bottomwin);
+	delwin(midwin);
+	delwin(footwin);
 #endif
 	/* Switch the cursor on, exit from curses, and restore terminal settings. */
 	restore_terminal();
@@ -291,57 +304,54 @@ void do_exit(void)
 	/* When unmodified, simply close.  Else, when doing automatic saving
 	 * and the file has a name, simply save.  Otherwise, ask the user. */
 	if (!openfile->modified)
-		choice = 0;
+		choice = NO;
 	else if (ISSET(SAVE_ON_EXIT) && openfile->filename[0] != '\0')
-		choice = 1;
+		choice = YES;
 	else {
 		if (ISSET(SAVE_ON_EXIT))
 			warn_and_briefly_pause(_("No file name"));
 
-		choice = do_yesno_prompt(FALSE, _("Save modified buffer? "));
+		choice = ask_user(YESORNO, _("Save modified buffer? "));
 	}
 
 	/* When not saving, or the save succeeds, close the buffer. */
-	if (choice == 0 || (choice == 1 && do_writeout(TRUE, TRUE) > 0))
+	if (choice == NO || (choice == YES && write_it_out(TRUE, TRUE) > 0))
 		close_and_go();
-	else if (choice != 1)
+	else if (choice != YES)
 		statusbar(_("Cancelled"));
 }
 
-/* Save the current buffer under the given name (or under the name "nano"
- * for a nameless buffer).  If needed, the name is modified to be unique. */
-void emergency_save(const char *plainname)
+/* Save the current buffer under the given name (or "nano.<pid>" when nameless)
+ * with suffix ".save".  If needed, the name is further suffixed to be unique. */
+void emergency_save(const char *filename)
 {
-	bool failed = TRUE;
-	char *targetname;
+	char *plainname, *targetname;
 
-	if (*plainname == '\0')
-		plainname = "nano";
+	if (*filename == '\0') {
+		plainname = nmalloc(28);
+		sprintf(plainname, "nano.%u", getpid());
+	} else
+		plainname = copy_of(filename);
 
 	targetname = get_next_filename(plainname, ".save");
 
-	if (*targetname != '\0')
-		failed = !write_file(targetname, NULL, TRUE, OVERWRITE, FALSE);
-
-	if (!failed)
+	if (*targetname == '\0')
+		fprintf(stderr, _("\nToo many .save files\n"));
+	else if (write_file(targetname, NULL, SPECIAL, OVERWRITE, NONOTES)) {
 		fprintf(stderr, _("\nBuffer written to %s\n"), targetname);
-	else if (*targetname != '\0')
-		fprintf(stderr, _("\nBuffer not written to %s: %s\n"),
-										targetname, strerror(errno));
-	else
-		fprintf(stderr, _("\nToo many .save files"));
-
-#ifndef NANO_TINY
-	/* Try to chmod/chown the saved file to the values of the original file,
-	 * but ignore any failure as we are in a hurry to get out. */
-	if (openfile->statinfo) {
-		IGNORE_CALL_RESULT(chmod(targetname, openfile->statinfo->st_mode));
-		IGNORE_CALL_RESULT(chown(targetname, openfile->statinfo->st_uid,
-												openfile->statinfo->st_gid));
-	}
+#if !defined(NANO_TINY) && defined(HAVE_CHMOD) && defined(HAVE_CHOWN)
+		/* Try to chmod/chown the saved file to the values of the original file,
+		 * but ignore any failure as we are in a hurry to get out. */
+		if (openfile->statinfo) {
+			IGNORE_CALL_RESULT(chmod(targetname, openfile->statinfo->st_mode));
+			IGNORE_CALL_RESULT(chown(targetname, openfile->statinfo->st_uid,
+													openfile->statinfo->st_gid));
+		}
 #endif
+	}
 
 	free(targetname);
+	free(plainname);
 }
 
 /* Die gracefully -- by restoring the terminal state and saving any buffers
@@ -350,6 +360,11 @@ void die(const char *msg, ...)
 {
 	va_list ap;
 	openfilestruct *firstone = openfile;
+	static int stabs = 0;
+
+	/* When dying for a second time, just give up. */
+	if (++stabs > 1)
+		exit(11);
 
 	restore_terminal();
 
@@ -364,7 +379,7 @@ void die(const char *msg, ...)
 
 	while (openfile) {
 #ifndef NANO_TINY
-		/* If the current buffer has a lockfile, remove it. */
+		/* If the current buffer has a lock file, remove it. */
 		if (openfile->lock_filename)
 			delete_lockfile(openfile->lock_filename);
 #endif
@@ -388,46 +403,46 @@ void die(const char *msg, ...)
 void window_init(void)
 {
 	/* When resizing, first delete the existing windows. */
-	if (edit != NULL) {
+	if (midwin != NULL) {
 		if (topwin != NULL)
 			delwin(topwin);
-		delwin(edit);
-		delwin(bottomwin);
+		delwin(midwin);
+		delwin(footwin);
 	}
 
 	topwin = NULL;
 
 	/* If the terminal is very flat, don't set up a title bar. */
 	if (LINES < 3) {
-		editwinrows = 1;
+		editwinrows = (ISSET(ZERO) ? LINES : 1);
 		/* Set up two subwindows.  If the terminal is just one line,
 		 * edit window and status-bar window will cover each other. */
-		edit = newwin(1, COLS, 0, 0);
-		bottomwin = newwin(1, COLS, LINES - 1, 0);
+		midwin = newwin(editwinrows, COLS, 0, 0);
+		footwin = newwin(1, COLS, LINES - 1, 0);
 	} else {
-		int toprows = ((ISSET(EMPTY_LINE) && LINES > 5) ? 2 : 1);
-		int bottomrows = ((ISSET(NO_HELP) || LINES < 5) ? 1 : 3);
+		int toprows = ((ISSET(EMPTY_LINE) && LINES > 6) ? 2 : 1);
+		int bottomrows = ((ISSET(NO_HELP) || LINES < 6) ? 1 : 3);
 
 #ifndef NANO_TINY
-		if (ISSET(MINIBAR))
+		if (ISSET(MINIBAR) || ISSET(ZERO))
 			toprows = 0;
 #endif
-		editwinrows = LINES - toprows - bottomrows;
+		editwinrows = LINES - toprows - bottomrows + (ISSET(ZERO) ? 1 : 0);
 
 		/* Set up the normal three subwindows. */
 		if (toprows > 0)
 			topwin = newwin(toprows, COLS, 0, 0);
-		edit = newwin(editwinrows, COLS, toprows, 0);
-		bottomwin = newwin(bottomrows, COLS, toprows + editwinrows, 0);
+		midwin = newwin(editwinrows, COLS, toprows, 0);
+		footwin = newwin(bottomrows, COLS, LINES - bottomrows, 0);
 	}
 
 	/* In case the terminal shrunk, make sure the status line is clear. */
-	wipe_statusbar();
+	wnoutrefresh(footwin);
 
 	/* When not disabled, turn escape-sequence translation on. */
 	if (!ISSET(RAW_SEQUENCES)) {
-		keypad(edit, TRUE);
-		keypad(bottomwin, TRUE);
+		keypad(midwin, TRUE);
+		keypad(footwin, TRUE);
 	}
 
 #ifdef ENABLED_WRAPORJUSTIFY
@@ -517,7 +532,7 @@ void usage(void)
 #ifdef ENABLE_HISTORIES
 	if (!ISSET(RESTRICTED))
 		print_opt("-H", "--historylog",
-					N_("Log & read search/replace string history"));
+					N_("Save & reload old search/replace strings"));
 #endif
 #ifdef ENABLE_NANORC
 	print_opt("-I", "--ignorercfiles", N_("Don't look at nanorc files"));
@@ -545,7 +560,7 @@ void usage(void)
 #ifdef ENABLE_HISTORIES
 	if (!ISSET(RESTRICTED))
 		print_opt("-P", "--positionlog",
-					N_("Log & read location of cursor position"));
+					N_("Save & restore position of the cursor"));
 #endif
 #ifdef ENABLE_JUSTIFY
 	print_opt(_("-Q <regex>"), _("--quotestr=<regex>"),
@@ -636,12 +651,9 @@ void usage(void)
 	print_opt("-x", "--nohelp", N_("Don't show the two help lines"));
 #ifndef NANO_TINY
 	print_opt("-y", "--afterends", N_("Make Ctrl+Right stop at word ends"));
-#endif
-	if (!ISSET(RESTRICTED))
-		print_opt("-z", "--suspendable", N_("Enable suspension"));
-#ifndef NANO_TINY
 	print_opt("-%", "--stateflags", N_("Show some states on the title bar"));
 	print_opt("-_", "--minibar", N_("Show a feedback bar at the bottom"));
+	print_opt("-0", "--zero", N_("Hide all bars, use whole terminal"));
 #endif
 #ifdef HAVE_LIBMAGIC
 	print_opt("-!", "--magic", N_("Also try magic to determine syntax"));
@@ -658,8 +670,8 @@ void version(void)
 	printf(_(" GNU nano, version %s\n"), VERSION);
 #endif
 #ifndef NANO_TINY
-	printf(" (C) 1999-2011, 2013-2021 Free Software Foundation, Inc.\n");
-	printf(_(" (C) 2014-%s the contributors to nano\n"), "2021");
+	printf(" (C) 1999-2011, 2013-2022 Free Software Foundation, Inc.\n");
+	printf(_(" (C) 2014-%s the contributors to nano\n"), "2022");
 #endif
 	printf(_(" Compiled options:"));
 
@@ -882,20 +894,21 @@ void signal_init(void)
 	sigaction(SIGTERM, &deed, NULL);
 
 #ifndef NANO_TINY
+#ifdef SIGWINCH
 	/* Trap SIGWINCH because we want to handle window resizes. */
 	deed.sa_handler = handle_sigwinch;
 	sigaction(SIGWINCH, &deed, NULL);
 #endif
-
-	/* In the suspend and continue handlers, block all other signals.
-	 * If we don't do this, other stuff interrupts them! */
-	sigfillset(&deed.sa_mask);
 #ifdef SIGTSTP
-	deed.sa_handler = do_suspend;
+	/* Prevent the suspend handler from getting interrupted. */
+	sigfillset(&deed.sa_mask);
+	deed.sa_handler = suspend_nano;
 	sigaction(SIGTSTP, &deed, NULL);
 #endif
+#endif /* !NANO_TINY */
 #ifdef SIGCONT
-	deed.sa_handler = do_continue;
+	sigfillset(&deed.sa_mask);
+	deed.sa_handler = continue_nano;
 	sigaction(SIGCONT, &deed, NULL);
 #endif
 
@@ -926,8 +939,9 @@ void handle_crash(int signal)
 }
 #endif
 
+#ifndef NANO_TINY
 /* Handler for SIGTSTP (suspend). */
-void do_suspend(int signal)
+void suspend_nano(int signal)
 {
 #ifdef ENABLE_MOUSE
 	disable_mouse_support();
@@ -949,20 +963,20 @@ void do_suspend(int signal)
 #endif
 }
 
-/* Put nano to sleep (if suspension is enabled). */
-void do_suspend_void(void)
+/* When permitted, put nano to sleep. */
+void do_suspend(void)
 {
-	if (!ISSET(SUSPENDABLE)) {
-		statusline(AHEM, _("Suspension is not enabled"));
-		beep();
-	} else
-		do_suspend(0);
+	if (in_restricted_mode())
+		return;
+
+	suspend_nano(0);
 
 	ran_a_tool = TRUE;
 }
+#endif /* !NANO_TINY */
 
 /* Handler for SIGCONT (continue after suspend). */
-void do_continue(int signal)
+void continue_nano(int signal)
 {
 #ifdef ENABLE_MOUSE
 	if (ISSET(USE_MOUSE))
@@ -978,18 +992,20 @@ void do_continue(int signal)
 #endif
 
 	/* Insert a fake keystroke, to neutralize a key-eating issue. */
-	ungetch(KEY_FLUSH);
+	ungetch(KEY_FRESH);
 }
 
 #if !defined(NANO_TINY) || defined(ENABLE_SPELLER) || defined(ENABLE_COLOR)
 /* Block or unblock the SIGWINCH signal, depending on the blockit parameter. */
 void block_sigwinch(bool blockit)
 {
+#ifdef SIGWINCH
 	sigset_t winch;
 
 	sigemptyset(&winch);
 	sigaddset(&winch, SIGWINCH);
 	sigprocmask(blockit ? SIG_BLOCK : SIG_UNBLOCK, &winch, NULL);
+#endif
 
 #ifndef NANO_TINY
 	if (the_window_resized)
@@ -1006,49 +1022,22 @@ void handle_sigwinch(int signal)
 	the_window_resized = TRUE;
 }
 
-/* Compute and store how many extra rows each line needs when softwrapping. */
-void compute_the_extra_rows_per_line_from(linestruct *fromline)
-{
-	if (ISSET(SOFTWRAP))
-		for (linestruct *line = fromline; line != NULL; line = line->next)
-			line->extrarows = extra_chunks_in(line);
-}
-
 /* Reinitialize and redraw the screen completely. */
 void regenerate_screen(void)
 {
-	const char *tty = ttyname(0);
-	int fd, result = 0;
-	struct winsize win;
-
 	/* Reset the trigger. */
 	the_window_resized = FALSE;
 
-	if (tty == NULL)
-		return;
-	fd = open(tty, O_RDWR);
-	if (fd == -1)
-		return;
-	result = ioctl(fd, TIOCGWINSZ, &win);
-	close(fd);
-	if (result == -1)
-		return;
+	/* Leave and immediately reenter curses mode, so that ncurses notices
+	 * the new screen dimensions and sets LINES and COLS accordingly. */
+	endwin();
+	refresh();
 
-	/* We could check whether COLS or LINES changed, and return otherwise,
-	 * but it seems curses does not always update these global variables. */
-#ifdef REDEFINING_MACROS_OK
-	COLS = win.ws_col;
-	LINES = win.ws_row;
-#endif
 	thebar = (ISSET(INDICATOR) && LINES > 5 && COLS > 9) ? 1 : 0;
 	bardata = nrealloc(bardata, LINES * sizeof(int));
 
 	editwincols = COLS - margin - thebar;
 
-	/* Do as the website suggests: leave and immediately reenter curses mode. */
-	endwin();
-	doupdate();
-
 	/* Put the terminal in the desired state again, and
 	 * recreate the subwindows with their (new) sizes. */
 	terminal_init();
@@ -1056,35 +1045,42 @@ void regenerate_screen(void)
 
 	/* If we have an open buffer, redraw the contents of the subwindows. */
 	if (openfile) {
-		compute_the_extra_rows_per_line_from(openfile->filetop);
 		ensure_firstcolumn_is_aligned();
 		draw_all_subwindows();
 	}
 }
 
-/* Handle the global toggle specified in flag. */
-void do_toggle(int flag)
+/* Invert the given global flag and adjust things for its new value. */
+void toggle_this(int flag)
 {
-	if (flag == SUSPENDABLE && in_restricted_mode())
-		return;
+	bool enabled = !ISSET(flag);
 
 	TOGGLE(flag);
 	focusing = FALSE;
 
 	switch (flag) {
+		case ZERO:
+			window_init();
+			draw_all_subwindows();
+			return;
 		case NO_HELP:
+			if (LINES < 6) {
+				statusline(AHEM, _("Too tiny"));
+				TOGGLE(flag);
+				return;
+			}
 			window_init();
 			draw_all_subwindows();
 			break;
-#ifdef ENABLE_MOUSE
-		case USE_MOUSE:
-			mouse_init();
-			break;
-#endif
+		case CONSTANT_SHOW:
+			if (ISSET(ZERO) || LINES == 1) {
+				statusline(AHEM, _("Not possible"));
+				TOGGLE(flag);
+			} else if (!ISSET(MINIBAR))
+				wipe_statusbar();
+			return;
 		case SOFTWRAP:
-			if (ISSET(SOFTWRAP))
-				compute_the_extra_rows_per_line_from(openfile->filetop);
-			else
+			if (!ISSET(SOFTWRAP))
 				openfile->firstcolumn = 0;
 			refresh_needed = TRUE;
 			break;
@@ -1097,30 +1093,37 @@ void do_toggle(int flag)
 			precalc_multicolorinfo();
 			refresh_needed = TRUE;
 			break;
+		case TABS_TO_SPACES:
+			if (openfile->syntax && openfile->syntax->tab) {
+				statusline(AHEM, _("Current syntax determines Tab"));
+				TOGGLE(flag);
+				return;
+			}
+			break;
+#endif
+#ifdef ENABLE_MOUSE
+		case USE_MOUSE:
+			mouse_init();
+			break;
 #endif
 	}
 
-	if (!ISSET(MINIBAR) && ISSET(STATEFLAGS))
-		if (flag == AUTOINDENT || flag == BREAK_LONG_LINES || flag == SOFTWRAP)
+	if (flag == AUTOINDENT || flag == BREAK_LONG_LINES || flag == SOFTWRAP) {
+		if (ISSET(MINIBAR) && !ISSET(ZERO) && ISSET(STATEFLAGS))
+			return;
+		if (ISSET(STATEFLAGS))
 			titlebar(NULL);
-
-	if (ISSET(MINIBAR) && (flag == NO_HELP || flag == LINE_NUMBERS ))
-		return;
-
-	if (flag == CONSTANT_SHOW)
-		wipe_statusbar();
-	else if (!ISSET(MINIBAR) || !ISSET(STATEFLAGS) || flag == SMART_HOME ||
-						flag == NO_SYNTAX || flag == WHITESPACE_DISPLAY ||
-						flag == CUT_FROM_CURSOR || flag == TABS_TO_SPACES ||
-						flag == USE_MOUSE || flag == SUSPENDABLE) {
-		bool enabled = ISSET(flag);
-
-		if (flag == NO_HELP || flag == NO_SYNTAX)
-			enabled = !enabled;
-
-		statusline(REMARK, "%s %s", _(flagtostr(flag)),
-									enabled ? _("enabled") : _("disabled"));
 	}
+
+	if (flag == NO_HELP || flag == LINE_NUMBERS || flag == WHITESPACE_DISPLAY)
+		if (ISSET(MINIBAR) || ISSET(ZERO) || LINES == 1)
+			return;
+
+	if (flag == NO_HELP || flag == NO_SYNTAX)
+		enabled = !enabled;
+
+	statusline(REMARK, "%s %s", _(epithet_of_flag(flag)),
+									enabled ? _("enabled") : _("disabled"));
 }
 #endif /* !NANO_TINY */
 
@@ -1246,12 +1249,11 @@ void confirm_margin(void)
 		editwincols = COLS - margin - thebar;
 
 #ifndef NANO_TINY
-		/* Recompute the softwrapped chunks for each line in the buffer,
-		 * and ensure a proper starting column for the first screen row. */
-		compute_the_extra_rows_per_line_from(openfile->filetop);
+		/* Ensure a proper starting column for the first screen row. */
 		ensure_firstcolumn_is_aligned();
-		focusing = keep_focus;
 #endif
+		focusing = keep_focus;
+
 		/* The margin has changed -- schedule a full refresh. */
 		refresh_needed = TRUE;
 	}
@@ -1265,6 +1267,11 @@ void unbound_key(int code)
 		/* TRANSLATORS: This refers to a sequence of escape codes
 		 * (from the keyboard) that nano does not recognize. */
 		statusline(AHEM, _("Unknown sequence"));
+#ifndef NANO_TINY
+	else if (code > KEY_F0 && code < KEY_F0 + 25)
+		/* TRANSLATORS: This refers to an unbound function key. */
+		statusline(AHEM, _("Unbound key: F%i"), code - KEY_F0);
+#endif
 	else if (code > 0x7F)
 		statusline(AHEM, _("Unbound key"));
 	else if (meta_key) {
@@ -1302,7 +1309,7 @@ int do_mouse(void)
 		return retval;
 
 	/* If the click was in the edit window, put the cursor in that spot. */
-	if (wmouse_trafo(edit, &click_row, &click_col, FALSE)) {
+	if (wmouse_trafo(midwin, &click_row, &click_col, FALSE)) {
 		linestruct *current_save = openfile->current;
 		ssize_t row_count = click_row - openfile->current_y;
 		size_t leftedge;
@@ -1385,7 +1392,7 @@ void suck_up_input_and_paste_it(void)
 	cutbuffer = line;
 
 	while (bracketed_paste) {
-		int input = get_kbinput(edit, BLIND);
+		int input = get_kbinput(midwin, BLIND);
 
 		if (input == '\r' || input == '\n') {
 			line->next = make_new_node(line);
@@ -1414,12 +1421,13 @@ void inject(char *burst, size_t count)
 	linestruct *thisline = openfile->current;
 	size_t datalen = strlen(thisline->data);
 #ifndef NANO_TINY
-	size_t old_amount = openfile->current->extrarows;
 	size_t original_row = 0;
+	size_t old_amount = 0;
 
 	if (ISSET(SOFTWRAP)) {
 		if (openfile->current_y == editwinrows - 1)
 			original_row = chunk_for(xplustabs(), thisline);
+		old_amount = extra_chunks_in(thisline);
 	}
 #endif
 
@@ -1474,23 +1482,19 @@ void inject(char *burst, size_t count)
 #endif
 
 #ifdef ENABLE_WRAPPING
-	/* Wrap the line when needed, and if so, schedule a refresh. */
-	if (ISSET(BREAK_LONG_LINES) && do_wrap())
-		refresh_needed = TRUE;
+	if (ISSET(BREAK_LONG_LINES))
+		do_wrap();
 #endif
 
 #ifndef NANO_TINY
 	/* When softwrapping and the number of chunks in the current line changed,
 	 * or we were on the last row of the edit window and moved to a new chunk,
 	 * we need a full refresh. */
-	if (ISSET(SOFTWRAP)) {
-		openfile->current->extrarows = extra_chunks_in(openfile->current);
-		if (openfile->current->extrarows != old_amount ||
+	if (ISSET(SOFTWRAP) && (extra_chunks_in(openfile->current) != old_amount ||
 					(openfile->current_y == editwinrows - 1 &&
-					chunk_for(xplustabs(), openfile->current) > original_row)) {
-			refresh_needed = TRUE;
-			focusing = FALSE;
-		}
+					chunk_for(xplustabs(), openfile->current) > original_row))) {
+		refresh_needed = TRUE;
+		focusing = FALSE;
 	}
 #endif
 
@@ -1520,10 +1524,9 @@ void process_a_keystroke(void)
 	const keystruct *shortcut;
 
 	/* Read in a keystroke, and show the cursor while waiting. */
-	input = get_kbinput(edit, VISIBLE);
+	input = get_kbinput(midwin, VISIBLE);
 
 	lastmessage = VACUUM;
-	hide_cursor = FALSE;
 
 #ifndef NANO_TINY
 	if (input == KEY_WINCH)
@@ -1534,7 +1537,7 @@ void process_a_keystroke(void)
 		/* If the user clicked on a shortcut, read in the key code that it was
 		 * converted into.  Else the click has been handled or was invalid. */
 		if (do_mouse() == 1)
-			input = get_kbinput(edit, BLIND);
+			input = get_kbinput(midwin, BLIND);
 		else
 			return;
 	}
@@ -1564,7 +1567,7 @@ void process_a_keystroke(void)
 
 	/* If we have a command, or if there aren't any other key codes waiting,
 	 * it's time to insert the gathered bytes into the edit buffer. */
-	if ((shortcut || get_key_buffer_len() == 0) && puddle != NULL) {
+	if ((shortcut || waiting_keycodes() == 0) && puddle != NULL) {
 		puddle[depth] = '\0';
 
 		inject(puddle, depth);
@@ -1611,8 +1614,8 @@ void process_a_keystroke(void)
 	}
 #endif
 #ifndef NANO_TINY
-	if (shortcut->func == do_toggle_void) {
-		do_toggle(shortcut->toggle);
+	if (shortcut->func == do_toggle) {
+		toggle_this(shortcut->toggle);
 		if (shortcut->toggle == CUT_FROM_CURSOR)
 			keep_cutbuffer = FALSE;
 		return;
@@ -1739,7 +1742,7 @@ int main(int argc, char **argv)
 		{"nowrap", 0, NULL, 'w'},
 #endif
 		{"nohelp", 0, NULL, 'x'},
-		{"suspendable", 0, NULL, 'z'},
+		{"suspendable", 0, NULL, 'z'},  /* Obsolete; remove in 2022. */
 #ifndef NANO_TINY
 		{"smarthome", 0, NULL, 'A'},
 		{"backup", 0, NULL, 'B'},
@@ -1768,6 +1771,7 @@ int main(int argc, char **argv)
 		{"afterends", 0, NULL, 'y'},
 		{"stateflags", 0, NULL, '%'},
 		{"minibar", 0, NULL, '_'},
+		{"zero", 0, NULL, '0'},
 #endif
 #ifdef HAVE_LIBMAGIC
 		{"magic", 0, NULL, '!'},
@@ -1785,10 +1789,12 @@ int main(int argc, char **argv)
 	/* Back up the terminal settings so that they can be restored. */
 	tcgetattr(STDIN_FILENO, &original_state);
 
+#if defined(F_GETFL) && defined(F_SETFL)
 	/* Get the state of standard input and ensure it uses blocking mode. */
 	stdin_flags = fcntl(STDIN_FILENO, F_GETFL, 0);
 	if (stdin_flags != -1)
 		fcntl(STDIN_FILENO, F_SETFL, stdin_flags & ~O_NONBLOCK);
+#endif
 
 #ifdef ENABLE_UTF8
 	/* If setting the locale is successful and it uses UTF-8, we will
@@ -1804,24 +1810,14 @@ int main(int argc, char **argv)
 	textdomain(PACKAGE);
 #endif
 
-#if defined(ENABLE_UTF8) && !defined(NANO_TINY)
-	if (MB_CUR_MAX > MAXCHARLEN)
-		fprintf(stderr, "Unexpected large character size: %i bytes"
-						" -- please report a bug\n", (int)MB_CUR_MAX);
-#endif
-
-	/* Set sensible defaults, different from what Pico does. */
+	/* Set a sensible default, different from what Pico does. */
 	SET(NO_WRAP);
-	SET(SMOOTH_SCROLL);
-
-	/* Give a small visual hint that nano has changed. */
-	SET(MORE_SPACE);
 
 	/* If the executable's name starts with 'r', activate restricted mode. */
 	if (*(tail(argv[0])) == 'r')
 		SET(RESTRICTED);
 
-	while ((optchr = getopt_long(argc, argv, "ABC:DEFGHIJ:KLMNOPQ:RST:UVWX:Y:Z"
+	while ((optchr = getopt_long(argc, argv, "0ABC:DEFGHIJ:KLMNOPQ:RST:UVWX:Y:Z"
 				"abcdef:ghijklmno:pqr:s:tuvwxyz$%_!", long_options, NULL)) != -1) {
 		switch (optchr) {
 #ifndef NANO_TINY
@@ -2056,7 +2052,6 @@ int main(int argc, char **argv)
 				break;
 #endif
 			case 'z':
-				SET(SUSPENDABLE);
 				break;
 #ifndef NANO_TINY
 			case '%':
@@ -2065,6 +2060,9 @@ int main(int argc, char **argv)
 			case '_':
 				SET(MINIBAR);
 				break;
+			case '0':
+				SET(ZERO);
+				break;
 #endif
 #ifdef HAVE_LIBMAGIC
 			case '!':
@@ -2077,6 +2075,10 @@ int main(int argc, char **argv)
 		}
 	}
 
+	/* Curses needs TERM; if it is unset, try falling back to a VT220. */
+	if (getenv("TERM") == NULL)
+		putenv("TERM=vt220");
+
 	/* Enter into curses mode.  Abort if this fails. */
 	if (initscr() == NULL)
 		exit(1);
@@ -2085,6 +2087,9 @@ int main(int argc, char **argv)
 	/* If the terminal can do colors, tell ncurses to switch them on. */
 	if (has_colors())
 		start_color();
+
+	/* When requested, suppress the default spotlight and error colors. */
+	rescind_colors = (getenv("NO_COLOR") != NULL);
 #endif
 
 	/* Set up the function and shortcut lists.  This needs to be done
@@ -2175,13 +2180,9 @@ int main(int argc, char **argv)
 			memmove(alt_speller, alt_speller + 1, strlen(alt_speller));
 #endif
 
-		/* If an rcfile undid the default settings, copy it to the new flags. */
+		/* If an rcfile undid the default setting, copy it to the new flag. */
 		if (!ISSET(NO_WRAP))
 			SET(BREAK_LONG_LINES);
-		if (!ISSET(SMOOTH_SCROLL))
-			SET(JUMPY_SCROLLING);
-		if (!ISSET(MORE_SPACE))
-			SET(EMPTY_LINE);
 
 		/* Simply OR the boolean flags from rcfile and command line. */
 		for (size_t i = 0; i < sizeof(flags) / sizeof(flags[0]); i++)
@@ -2200,11 +2201,10 @@ int main(int argc, char **argv)
 	if (ISSET(BOLD_TEXT))
 		hilite_attribute = A_BOLD;
 
-	/* When in restricted mode, disable backups, suspending, and history files,
-	 * since they allow writing to files not specified on the command line. */
+	/* When in restricted mode, disable backups and history files, since they
+	 * would allow writing to files not specified on the command line. */
 	if (ISSET(RESTRICTED)) {
 		UNSET(MAKE_BACKUP);
-		UNSET(SUSPENDABLE);
 #ifdef ENABLE_NANORC
 		UNSET(HISTORYLOG);
 		UNSET(POSITIONLOG);
@@ -2215,6 +2215,10 @@ int main(int argc, char **argv)
 	if (ISSET(RAW_SEQUENCES))
 		UNSET(USE_MOUSE);
 
+	/* When suppressing title bar or minibar, suppress also the help lines. */
+	if (ISSET(ZERO))
+		SET(NO_HELP);
+
 #ifdef ENABLE_HISTORIES
 	/* Initialize the pointers for the Search/Replace/Execute histories. */
 	history_init();
@@ -2231,7 +2235,7 @@ int main(int argc, char **argv)
 		load_history();
 	if (ISSET(POSITIONLOG))
 		load_poshistory();
-#endif /* ENABLE_HISTORIES */
+#endif
 
 #ifndef NANO_TINY
 	/* If a backup directory was specified and we're not in restricted mode,
@@ -2265,7 +2269,7 @@ int main(int argc, char **argv)
 		die(_("Bad quoting regex \"%s\": %s\n"), quotestr, message);
 	} else
 		free(quotestr);
-#endif /* ENABLE_JUSTIFY */
+#endif
 
 #ifdef ENABLE_SPELLER
 	/* If we don't have an alternative spell checker after reading the
@@ -2326,6 +2330,7 @@ int main(int argc, char **argv)
 		interface_color_pair[SCROLL_BAR] = A_NORMAL;
 		interface_color_pair[SELECTED_TEXT] = hilite_attribute;
 		interface_color_pair[SPOTLIGHTED] = A_REVERSE;
+		interface_color_pair[MINI_INFOBAR] = hilite_attribute;
 		interface_color_pair[PROMPT_BAR] = hilite_attribute;
 		interface_color_pair[STATUS_BAR] = hilite_attribute;
 		interface_color_pair[ERROR_MESSAGE] = hilite_attribute;
@@ -2406,9 +2411,9 @@ int main(int argc, char **argv)
 #endif
 		/* If there's a +LINE[,COLUMN] argument here, eat it up. */
 		if (optind < argc - 1 && argv[optind][0] == '+') {
+#ifndef NANO_TINY
 			int n = 1;
 
-#ifndef NANO_TINY
 			while (isalpha(argv[optind][n])) {
 				switch (argv[optind][n++]) {
 					case 'c': SET(CASE_SENSITIVE); break;
@@ -2431,7 +2436,11 @@ int main(int argc, char **argv)
 				optind++;
 			} else
 #endif
-			if (!parse_line_column(&argv[optind++][n], &givenline, &givencol))
+			/* When there is nothing after the "+", understand it as go-to-EOF,
+			 * otherwise parse and store the given number(s).*/
+			if (argv[optind++][1] == '\0')
+				givenline = -1;
+			else if (!parse_line_column(&argv[optind - 1][1], &givenline, &givencol))
 				statusline(ALERT, _("Invalid line or column number"));
 		}
 
@@ -2449,7 +2458,7 @@ int main(int argc, char **argv)
 
 		/* If a position was given on the command line, go there. */
 		if (givenline != 0 || givencol != 0)
-			do_gotolinecolumn(givenline, givencol, FALSE, FALSE);
+			goto_line_and_column(givenline, givencol, FALSE, FALSE);
 #ifndef NANO_TINY
 		else if (searchstring != NULL) {
 			if (ISSET(USE_REGEXP))
@@ -2457,7 +2466,7 @@ int main(int argc, char **argv)
 			if (!findnextstr(searchstring, FALSE, JUSTFIND, NULL,
 							ISSET(BACKWARDS_SEARCH), openfile->filetop, 0))
 				not_found_msg(searchstring);
-			else if (lastmessage == HUSH)
+			else if (lastmessage <= REMARK)
 				wipe_statusbar();
 			openfile->placewewant = xplustabs();
 			if (ISSET(USE_REGEXP))
@@ -2472,11 +2481,14 @@ int main(int argc, char **argv)
 			ssize_t savedline, savedcol;
 			/* If edited before, restore the last cursor position. */
 			if (has_old_position(argv[optind - 1], &savedline, &savedcol))
-				do_gotolinecolumn(savedline, savedcol, FALSE, FALSE);
+				goto_line_and_column(savedline, savedcol, FALSE, FALSE);
 		}
 #endif
 	}
 
+	/* After handling the files on the command line, allow inserting files. */
+	UNSET(NOREAD_MODE);
+
 	/* If no filenames were given, or all of them were invalid things like
 	 * directories, then open a blank buffer and allow editing.  Otherwise,
 	 * switch from the last opened file to the next, that is: the first. */
@@ -2502,11 +2514,16 @@ int main(int argc, char **argv)
 #ifdef ENABLE_NANORC
 	if (startup_problem != NULL)
 		statusline(ALERT, startup_problem);
+
+#define NOTREBOUND  first_sc_for(MMAIN, do_help) && \
+						first_sc_for(MMAIN, do_help)->keycode == 0x07
+#else
+#define NOTREBOUND  TRUE
 #endif
 
 #ifdef ENABLE_HELP
 	if (*openfile->filename == '\0' && openfile->totsize == 0 &&
-				openfile->next == openfile && !ISSET(NO_HELP))
+				openfile->next == openfile && !ISSET(NO_HELP) && NOTREBOUND)
 		statusbar(_("Welcome to nano.  For basic help, type Ctrl+G."));
 #endif
 
@@ -2517,36 +2534,45 @@ int main(int argc, char **argv)
 		confirm_margin();
 #endif
 #ifdef __linux__
-		if (on_a_vt && get_key_buffer_len() == 0)
+		if (on_a_vt && waiting_keycodes() == 0)
 			mute_modifiers = FALSE;
 #endif
 		if (currmenu != MMAIN)
 			bottombars(MMAIN);
 
 #ifndef NANO_TINY
-		if (ISSET(MINIBAR) && lastmessage < REMARK)
+		if (ISSET(MINIBAR) && !ISSET(ZERO) && LINES > 1 && lastmessage < REMARK)
 			minibar();
 		else
 #endif
 		/* Update the displayed current cursor position only when there
 		 * is no message and no keys are waiting in the input buffer. */
-		if (ISSET(CONSTANT_SHOW) && lastmessage == VACUUM && get_key_buffer_len() == 0)
+		if (ISSET(CONSTANT_SHOW) && lastmessage == VACUUM && LINES > 1 &&
+								!ISSET(ZERO) && waiting_keycodes() == 0)
 			report_cursor_position();
 
 		as_an_at = TRUE;
 
-		/* Refresh just the cursor position or the entire edit window. */
-		if (!refresh_needed) {
-			place_the_cursor();
-			wnoutrefresh(edit);
-		} else
+		if ((refresh_needed && LINES > 1) || (LINES == 1 && lastmessage <= HUSH))
 			edit_refresh();
+		else
+			place_the_cursor();
 
 #ifndef NANO_TINY
-		/* Let the next keystroke cancel the highlighting of a search match. */
-		refresh_needed = spotlighted;
-		spotlighted = FALSE;
+		/* In barless mode, either redraw a relevant status message,
+		 * or overwrite a minor, redundant one. */
+		if (ISSET(ZERO) && lastmessage > HUSH) {
+			if (openfile->current_y == editwinrows - 1 && LINES > 1) {
+				edit_scroll(FORWARD);
+				wnoutrefresh(midwin);
+			}
+			wredrawln(footwin, 0, 1);
+			wnoutrefresh(footwin);
+			place_the_cursor();
+		} else if (ISSET(ZERO) && lastmessage > VACUUM)
+			wredrawln(midwin, editwinrows - 1, 1);
 #endif
+
 		errno = 0;
 		focusing = TRUE;
 
